#include <netinet/in.h>
#include "mpglib.h"

#include "mpg123.h"
#include "getbits.h"

// one-time initialization of the decoder
void MPGLib::init() // static
{
    make_decode_tables(32768);
    init_layer3(32);
}

MPGLib::MPGLib()
{
    pcm_point = 0;

    memset(&mp, 0, sizeof(mp));
    memset(&fr, 0, sizeof(fr));

    memset(&bsbufend, 0, sizeof(bsbufend));
    bsbufold_end = 0;
    memset(bsspace, 0, sizeof(bsspace));
    bsbuf = bsspace[1] + 512;
    bsbufold = 0;
    bsnum = 0;

    fr.pCB = this;

    fr.down_sample_sblimit = 32;
    fr.single = -1;

    frame_num = 0;
    used_reservoir = 0;

    synth_reset();
}

int tabsel_123[2][3][16] = {
   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
     {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
     {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};

long freqs[9] = 
    { 44100, 48000, 32000, 22050, 24000, 16000 , 11025 , 12000 , 8000 };

/*
 * decode a header and write the information into the frame structure
 */
static int decode_header(struct frame *fr, unsigned long newhead)
{
    if( newhead & (1<<20) ) {
	fr->lsf = (newhead & (1<<19)) ? 0x0 : 0x1;
	fr->mpeg25 = 0;
    }
    else {
	fr->lsf = 1;
	fr->mpeg25 = 1;
    }

    /* 
     * CHECKME: should be add more consistency checks here ?  
     * changed layer, changed CRC bit, changed sampling frequency 
     */
    {
	fr->lay = 4-((newhead>>17)&3);
	if( ((newhead>>10)&0x3) == 0x3) {
	    fprintf(stderr,"Stream error\n");
	    return 0;
	}
	if(fr->mpeg25) {
	    fr->sampling_frequency = 6 + ((newhead>>10)&0x3);
	}
	else
	    fr->sampling_frequency = ((newhead>>10)&0x3) + (fr->lsf*3);
	fr->error_protection = ((newhead>>16)&0x1)^0x1;
    }

    fr->bitrate_index = ((newhead>>12)&0xf);
    fr->padding   = ((newhead>>9)&0x1);
    fr->extension = ((newhead>>8)&0x1);
    fr->mode      = ((newhead>>6)&0x3);
    fr->mode_ext  = ((newhead>>4)&0x3);
    fr->copyright = ((newhead>>3)&0x1);
    fr->original  = ((newhead>>2)&0x1);
    fr->emphasis  = newhead & 0x3;

    fr->stereo    = (fr->mode == MPG_MD_MONO) ? 1 : 2;

    switch(fr->lay) {
    case 3:
        if(fr->lsf)
	    fr->sideInfoSize = (fr->stereo == 1) ? 9 : 17;
        else
	    fr->sideInfoSize = (fr->stereo == 1) ? 17 : 32;
        if(fr->error_protection)
	    fr->sideInfoSize += 2;
        fr->framesize  = (long) tabsel_123[fr->lsf][2][fr->bitrate_index] * 
            144000;
        fr->framesize /= freqs[fr->sampling_frequency]<<(fr->lsf);
        fr->framesize = fr->framesize + fr->padding - 4;
        fr->padsize = fr->padding;
        break; 
    default:
        return (0);
    }

    if(!fr->bitrate_index)
        return (0);

    fr->thishead = newhead;

    return 1;
}

/* This is the worst case number of previous frames on which an MP3 frame
   can depend. */
#define R_MP3_RESERVOIR_WORST_CASE 9

unsigned int MPGLib::process_frame(const unsigned char* frame_data,
                                   unsigned int frame_len,
                                   short* pcm_data)
{
    unsigned long newhead;
    if (frame_len < sizeof(newhead))
        return 0;
    memcpy(&newhead, frame_data, sizeof(newhead));
    newhead = ntohl(newhead);

    if (!decode_header(&fr, newhead))
        return 0;

    if (frame_len - sizeof(newhead) != fr.framesize)
        return 0;

    /* flip/init buffer for Layer 3 */
    bsbufold = bsbuf;
    bsbufold_end = bsbufend[bsnum];
    bsbuf = bsspace[bsnum] + 512;
    bsnum = (bsnum + 1) & 1;
    bsbufend[bsnum] = fr.framesize;

    memcpy(bsbuf, &frame_data[4], fr.framesize);

    bsi.bitindex = 0;
    bsi.wordpointer = (unsigned char *) bsbuf;

    if (fr.error_protection) {
       /* skip crc, we are byte aligned here */
       getbyte(&bsi);
       getbyte(&bsi);
    }

    pcm_point = 0;
    pcm_sample = (unsigned char*)pcm_data;

    used_reservoir = 0;

    do_layer3(&mp, &fr);

    if (frame_num < R_MP3_RESERVOIR_WORST_CASE) {
        frame_num++;

        if (used_reservoir) {
            return 0;
        } else {
            /* play frames from now on */
            frame_num = R_MP3_RESERVOIR_WORST_CASE;
        }
    }

    return ((unsigned int)pcm_point) / (2 * sizeof(short));
}

void MPGLib::set_pointer(int ssize, long backstep)
{
    bsi.wordpointer = bsbuf + ssize - backstep;
    if (backstep) {
	memcpy(bsi.wordpointer, bsbufold + bsbufold_end - backstep, backstep);
        used_reservoir = 1;
    }
    bsi.bitindex = 0; 
}
